@@ -35,13 +35,12 @@ from grapheekdb.lib.exceptions import GrapheekInvalidDataTypeException
 from grapheekdb.lib.exceptions import GrapheekUnknownScriptException
 from grapheekdb.lib.exceptions import GrapheekUnknownAlias
 
-from grapheekdb.backends.data.filtertools import build_attr_filter_funcs
+from grapheekdb.backends.data.filtertools import build_filter_funcs
 from grapheekdb.backends.data.filtertools import filter_entities
 
-from grapheekdb.backends.data.indexes import ExactIndex
-from grapheekdb.backends.data.indexes import choose_index_or_scan
-
+from grapheekdb.backends.data.indexes import ExactIndex, normalize_value
 from grapheekdb.backends.data.operations import Addition, Removal
+from grapheekdb.backends.data.optimizer import Optimizer
 
 
 # misc iterator :
@@ -67,18 +66,6 @@ def check_and_get_count(args):
     return count
 
 
-def get_kind_ids(graph, txn, kind):
-    ENTITY_COUNTER = METADATA_VERTEX_COUNTER if kind == KIND_VERTEX else METADATA_EDGE_COUNTER
-    METADATA_ID_LIST_PREFIX = METADATA_VERTEX_ID_LIST_PREFIX if kind == KIND_VERTEX else METADATA_EDGE_ID_LIST_PREFIX
-    limit = int(graph._get(None, ENTITY_COUNTER)) / CHUNK_SIZE
-    keys = [build_key(METADATA_ID_LIST_PREFIX, i) for i in range(0, limit + 1)]
-    list_entity_ids = graph._bulk_get_lst(txn, keys)
-    for entity_ids in list_entity_ids:
-        if entity_ids != UNDEFINED:
-            for entity_id in entity_ids:
-                yield entity_id
-
-
 class DotExport(object):
     """
     This class is mainly intended to be used with an IPython Graphviz Magics extension
@@ -261,7 +248,7 @@ class EntityIterator(object):
         self._graph = graph
         self._src_kind = src_kind
         self._filters = filters
-        self._attr_filter_funcs = build_attr_filter_funcs(**filters)
+        self._filter_funcs = build_filter_funcs(**filters)
         self._id_iterator = id_iterator
         self._parent = parent
         self._neighbors_cache = neighbors_cache if neighbors_cache is not None else {}
@@ -278,7 +265,7 @@ class EntityIterator(object):
             self._neighbors_cache = {}
 
     def _iterate(self, on_item=False):
-        for item in filter_entities(self._graph, self._src_kind, self._id_iterator, self._attr_filter_funcs, on_item):
+        for item in filter_entities(self._graph, self._src_kind, self._id_iterator, self._filter_funcs, on_item):
             yield item
         self._finalize()
 
@@ -365,7 +352,7 @@ class EntityIterator(object):
                         if int(entity_id) != int(neighbor_node_id):
                             yield neighbor_node_id
 
-        filter_funcs = build_attr_filter_funcs(**filters)
+        filter_funcs = build_filter_funcs(**filters)
         if count == 1:
             return EntityIterator(
                 self._graph,
@@ -814,7 +801,7 @@ class Edge(Entity):
 
 class BaseGraph(GraphReadWrite):
 
-    def __del__(self):
+    def __del__(self):  # pragma : no cover
         # ensure connection is closed
         try:
             if not self._closed:
@@ -822,6 +809,39 @@ class BaseGraph(GraphReadWrite):
         except AttributeError:
             pass
 
+    def __initialize__(self):
+        # "Rebuilding" indexes :
+        txn = self._transaction_begin()
+        try:
+            # Get vertex and edge count (useful for queries where we need to compare index perf and seq scan) :
+            self._node_count = self._get(txn, METADATA_VERTEX_COUNTER) - self._get(txn, METADATA_VERTEX_REMOVED_COUNTER)
+            self._edge_count = self._get(txn, METADATA_EDGE_COUNTER) - self._get(txn, METADATA_EDGE_REMOVED_COUNTER)
+            # Instantiating vertex and edge indexes
+            self._node_indexes = []
+            self._edge_indexes = []
+            #   1st vertex indexes :
+            index_count = self._get(txn, METADATA_VERTEX_INDEX_COUNTER)
+            for index_id in range(0, index_count):
+                fields, filter_tuples = self._get(txn, build_key(METADATA_VERTEX_INDEX_FIELDS_PREFIX, index_id))
+                filters = dict(filter_tuples)
+                if fields != UNDEFINED:
+                    prefix = build_key(METADATA_VERTEX_INDEX_PREFIX, index_id)
+                    self._node_indexes.append(ExactIndex(KIND_VERTEX, prefix, *fields, **filters))
+            #   2nd edge indexes (~ same code, but I'm not sure it makes sense to factor...)
+            index_count = self._get(txn, METADATA_EDGE_INDEX_COUNTER)
+            for index_id in range(0, index_count):
+                fields, filter_tuples = self._get(txn, build_key(METADATA_VERTEX_INDEX_FIELDS_PREFIX, index_id))
+                filters = dict(filter_tuples)
+                if fields != UNDEFINED:
+                    prefix = build_key(METADATA_VERTEX_INDEX_PREFIX, index_id)
+                    self._edge_indexes.append(ExactIndex(KIND_VERTEX, prefix, *fields, **filters))
+            self._transaction_commit(txn)
+        except Exception, e:  # pragma : no cover
+            self._transaction_rollback(txn)
+            raise GrapheekDataException(repr(e))
+        # Instantiate an optimizer (which will choose between scan and sequential scan AND (in the future) cache best index for most runned queries)
+        self._optimizer = Optimizer(self)
+
     def _prepare_database(self, txn):
         self._set(txn, PREPARED, 1)
         self._set(txn, METADATA_VERTEX_COUNTER, 0)
@@ -840,32 +860,7 @@ class BaseGraph(GraphReadWrite):
             except Exception, e:
                 self._transaction_rollback(txn)
                 raise GrapheekDataPreparationFailedException(repr(e))
-        txn = self._transaction_begin()
-        try:
-            # Get vertex and edge count (useful for queries where we need to compare index perf and seq scan) :
-            self._node_count = self._get(txn, METADATA_VERTEX_COUNTER) - self._get(txn, METADATA_VERTEX_REMOVED_COUNTER)
-            self._edge_count = self._get(txn, METADATA_EDGE_COUNTER) - self._get(txn, METADATA_EDGE_REMOVED_COUNTER)
-            # Instantiating vertex and edge indexes
-            self._node_indexes = []
-            self._edge_indexes = []
-            #   1st vertex indexes :
-            index_count = self._get(txn, METADATA_VERTEX_INDEX_COUNTER)
-            for index_id in range(0, index_count):
-                fields = self._get(txn, build_key(METADATA_VERTEX_INDEX_FIELDS_PREFIX, index_id))
-                if fields != UNDEFINED:
-                    prefix = build_key(METADATA_VERTEX_INDEX_PREFIX, index_id)
-                    self._node_indexes.append(ExactIndex(KIND_VERTEX, prefix, *fields))
-            #   2nd edge indexes (~ same code, but I'm not sure it makes sense to factor...)
-            index_count = self._get(txn, METADATA_EDGE_INDEX_COUNTER)
-            for index_id in range(0, index_count):
-                fields = self._get(txn, build_key(METADATA_EDGE_INDEX_FIELDS_PREFIX, index_id))
-                if fields != UNDEFINED:
-                    prefix = build_key(METADATA_VERTEX_INDEX_PREFIX, index_id)
-                    self._edge_indexes.append(ExactIndex(KIND_VERTEX, prefix, *fields))
-            self._transaction_commit(txn)
-        except Exception, e:  # pragma : no cover
-            self._transaction_rollback(txn)
-            raise GrapheekDataException(repr(e))
+        self.__initialize__()
 
     # Following methods MUST be overriden by child classes :
 
@@ -1230,29 +1225,6 @@ class BaseGraph(GraphReadWrite):
         Klass = Node if kind == KIND_VERTEX else Edge
         return Klass(entity_id, self)
 
-    def __index_or_seq_scan_iterator(self, _kind, _seq_count, **filters):
-
-        def indexed_entity_generator(entity_ids):
-            for entity_id in entity_ids:
-                yield entity_id
-
-        iterator = None
-        indexes = self._node_indexes if _kind == KIND_VERTEX else self._edge_indexes
-
-        if filters and indexes:
-            # There's some filters, so it *may* be useful to use filters
-            # In order to know if it is useful, we will ask  every filter
-            # to estimate the number of entity ids that they could return
-            # we will, then, choose between the best index (given the filters)
-            # and sequential iterator
-            best_index = choose_index_or_scan(_seq_count, indexes, filters)
-            if best_index:
-                # Ok, an index will (hopefully :p) returns less ids than the seq scan, let's get ids :
-                entity_ids = best_index.ids(None, filters)
-                iterator = indexed_entity_generator(entity_ids)
-
-        return iterator
-
     # Traversals given entity ids :
 
     def _neighbors(self, _kind, _entity_id, _traversal, _cache=None, **filters):
@@ -1298,7 +1270,7 @@ class BaseGraph(GraphReadWrite):
                 # to challenge every indexed ids to check that it is among current entity neighbours
                 # so it will lead to
                 # which can cost a lot (<index_count> disk access for persistent backends)
-                iterator = self.__index_or_seq_scan_iterator(_kind, seq_count * 2, **filters)
+                iterator = self._optimizer.index_or_seq_scan_iterator(_kind, seq_count * 2, **filters)
             if iterator is None:
                 for entity_id in iter(neighbour_ids):
                     yield entity_id
@@ -1317,9 +1289,9 @@ class BaseGraph(GraphReadWrite):
         else:
             seq_count = self._edge_count
 
-        iterator = self.__index_or_seq_scan_iterator(_kind, seq_count, **filters)
+        iterator = self._optimizer.index_or_seq_scan_iterator(_kind, seq_count, **filters)
         if iterator is None:  # no index is better than sequential scan, so choose it
-            iterator = get_kind_ids(self, None, _kind)
+            iterator = self._optimizer.get_kind_ids(None, _kind)
 
         return EntityIterator(self, _kind, filters, iterator)
 
@@ -1339,10 +1311,10 @@ class BaseGraph(GraphReadWrite):
         check_valid_data(kwargs)
         return self._add_node(None, kwargs)
 
-    def bulk_add_node(self, node_defns):
+    def bulk_add_node(self, node_defns, silent=False):
         for data in node_defns:
             check_valid_data(data)
-        return self._bulk_add_node(node_defns)
+        return self._bulk_add_node(node_defns, silent)
 
     def add_edge(self, _source, _target, **kwargs):
         check_base_node(_source)
@@ -1351,7 +1323,7 @@ class BaseGraph(GraphReadWrite):
 
         return self._add_edge(None, _source, _target, kwargs)
 
-    def bulk_add_edge(self, edge_defns):
+    def bulk_add_edge(self, edge_defns, silent=False):
         try:
             for source, target, data in edge_defns:
                 check_base_node(source)
@@ -1359,10 +1331,10 @@ class BaseGraph(GraphReadWrite):
                 check_valid_data(data)
         except (TypeError, ValueError), e:
             raise GrapheekDataException(repr(e))
-        return self._bulk_add_edge(edge_defns)
+        return self._bulk_add_edge(edge_defns, silent)
 
-    def bulk_add_edge_by_ids(self, edge_id_defns):
-        return self._bulk_add_edge_by_ids(edge_id_defns)
+    def bulk_add_edge_by_ids(self, edge_id_defns, silent=False):
+        return self._bulk_add_edge_by_ids(edge_id_defns, silent)
 
     def add_edge_by_ids(self, _source_id, _target_id, **kwargs):
         return self._add_edge_by_ids(_source_id, _target_id, kwargs)
@@ -1380,11 +1352,15 @@ class BaseGraph(GraphReadWrite):
         for index in indexes:
             index.remove(txn, entity_id)
 
-    def _add_entity_index(self, kind, *args):
+    def _add_entity_index(self, _kind, *args, **filters):
         fields = list(args)
         fields.sort()
-        assert(kind in (KIND_VERTEX, KIND_EDGE))
-        if kind == KIND_VERTEX:
+        filter_tuples = filters.items()
+        filter_tuples.sort()
+        index_signature = [fields, filter_tuples]
+        index_signature_string = normalize_value(index_signature)
+        assert(_kind in (KIND_VERTEX, KIND_EDGE))
+        if _kind == KIND_VERTEX:
             METADATA_INDEX_PREFIX = METADATA_VERTEX_INDEX_PREFIX
             METADATA_INDEX_COUNTER = METADATA_VERTEX_INDEX_COUNTER
             METADATA_COUNTER = METADATA_VERTEX_COUNTER
@@ -1399,7 +1375,7 @@ class BaseGraph(GraphReadWrite):
             METADATA_INDEX_FIELDS_PREFIX = METADATA_EDGE_INDEX_FIELDS_PREFIX
             entity_indexes = self._edge_indexes
         # Checking if index already exists (won't create if it is the case) :
-        index_key = '/'.join([METADATA_INDEX_PREFIX] + fields)
+        index_key = '/'.join([METADATA_INDEX_PREFIX] + [index_signature_string])
         index_id = self._get(None, index_key)
         if index_id != UNDEFINED:
             # Index already exists
@@ -1409,16 +1385,16 @@ class BaseGraph(GraphReadWrite):
             index_id = self._new_id_for_key(txn, METADATA_INDEX_COUNTER)
             self._set(txn, index_key, index_id)
             self._append_to_lst(txn, METADATA_INDEX_LIST, index_id)  # "Registering" index in index list
-            self._set(txn, build_key(METADATA_INDEX_FIELDS_PREFIX, index_id), fields)
+            self._set(txn, build_key(METADATA_INDEX_FIELDS_PREFIX, index_id), index_signature)
             entity_count = self._get(txn, METADATA_COUNTER)
             if entity_count == UNDEFINED:
                 raise GrapheekDataException
-            prefix = build_key(KIND_INDEX, kind, index_id)
+            prefix = build_key(KIND_INDEX, _kind, index_id)
             # NOTE : So far, only ExactIndex is supported :
-            index = ExactIndex(self, kind, prefix, *fields)
+            index = ExactIndex(self, _kind, prefix, *fields, **filters)
             # Entity iterator : index will iterate on it to create denormalized data
             # that will (should ;)) speedup query
-            id_iterator = get_kind_ids(self, txn, kind)
+            id_iterator = self._optimizer.get_kind_ids(txn, _kind)
             # Building index :
             index.build(txn, id_iterator)
             # Everything seems to be ok, adding index to proper list
@@ -1428,23 +1404,27 @@ class BaseGraph(GraphReadWrite):
             self._transaction_rollback(txn)
             raise GrapheekIndexCreationFailedException(repr(e))
 
-    def add_node_index(self, *fields):
-        self._add_entity_index(KIND_VERTEX, *fields)
+    def add_node_index(self, *fields, **filters):
+        self._add_entity_index(KIND_VERTEX, *fields, **filters)
 
-    def add_edge_index(self, *fields):
-        self._add_entity_index(KIND_EDGE, *fields)
+    def add_edge_index(self, *fields, **filters):
+        self._add_entity_index(KIND_EDGE, *fields, **filters)
 
     def get_node_indexes(self):
-        return [index._fields for index in self._node_indexes]
+        return [[index._fields, dict(index._filter_items)] for index in self._node_indexes]
 
     def get_edge_indexes(self):
-        return [index._fields for index in self._edge_indexes]
+        return [[index._fields, dict(index._filter_items)] for index in self._edge_indexes]
 
-    def _remove_entity_index(self, kind, *args):
-        assert(kind in (KIND_VERTEX, KIND_EDGE))
+    def _remove_entity_index(self, _kind, *args, **kwargs):
+        assert(_kind in (KIND_VERTEX, KIND_EDGE))
         fields = list(args)
         fields.sort()
-        if kind == KIND_VERTEX:
+        filter_tuples = kwargs.items()
+        filter_tuples.sort()
+        index_signature = [fields, filter_tuples]
+        index_signature_string = normalize_value(index_signature)
+        if _kind == KIND_VERTEX:
             indexes = self._node_indexes
             METADATA_INDEX_PREFIX = METADATA_VERTEX_INDEX_PREFIX
             METADATA_INDEX_LIST = METADATA_VERTEX_INDEX_LIST
@@ -1456,7 +1436,7 @@ class BaseGraph(GraphReadWrite):
             METADATA_INDEX_FIELDS_PREFIX = METADATA_EDGE_INDEX_FIELDS_PREFIX
         index_to_remove = None
         for index in indexes:
-            if index._fields == fields:
+            if index._fields == fields and index._filter_items == set(kwargs.items()):
                 index_to_remove = index
                 break
         if index_to_remove is None:
@@ -1468,7 +1448,7 @@ class BaseGraph(GraphReadWrite):
             # Removing index from index list :
             indexes.remove(index)
             # Removing "management key"
-            index_key = '/'.join([METADATA_INDEX_PREFIX] + fields)
+            index_key = '/'.join([METADATA_INDEX_PREFIX] + [index_signature_string])
             index_id = self._get(txn, index_key)
             # Remove index key :
             self._remove(txn, index_key)
