@@ -4,10 +4,7 @@ from types import BooleanType, DictType, FloatType, IntType
 from types import ListType, LongType, NoneType, StringType, TupleType, UnicodeType
 
 import zmq.green as zmq
-try:
-    import ujson as json
-except ImportError:  # pragma : no cover
-    import json
+import msgpack
 
 from grapheekdb.backends.data.keys import KIND_VERTEX, KIND_EDGE
 
@@ -309,10 +306,10 @@ class ProxyGraph(GraphReadWrite):
         self._zmq_socket.connect(address)
 
     def _request(self, commands):
-        data = json.dumps(commands)
+        data = msgpack.dumps(commands)
         self._zmq_socket.send(data)
         raw = self._zmq_socket.recv()
-        msg = json.loads(raw)
+        msg = msgpack.loads(raw)
         return unmarshall(self, msg)
 
     def v(self, entity_id):
@@ -334,10 +331,10 @@ class ProxyGraph(GraphReadWrite):
         command = ['add_node', [], kwargs]
         return self._request([command])
 
-    def bulk_add_node(self, node_defns):
+    def bulk_add_node(self, node_defns, silent=False):
         for data in node_defns:
             check_valid_data(data)
-        command = ['bulk_add_node', [node_defns], {}]
+        command = ['bulk_add_node', [node_defns], {'silent': silent}]
         return self._request([command])
 
     def add_edge(self, _source, _target, **kwargs):
@@ -349,7 +346,7 @@ class ProxyGraph(GraphReadWrite):
         command = ['add_edge_by_ids', [source_id, target_id], kwargs]
         return self._request([command])
 
-    def bulk_add_edge(self, edge_defns):
+    def bulk_add_edge(self, edge_defns, silent=False):
         try:
             for source, target, data in edge_defns:
                 check_proxy_node(source)
@@ -360,19 +357,19 @@ class ProxyGraph(GraphReadWrite):
         # Don't be suprised by [[ ... ]] instead of [ ... ]
         # Server function is expecting only one argument :
         args = [[(source.get_id(), target.get_id(), data) for source, target, data in edge_defns]]
-        command = ['bulk_add_edge_by_ids', args, {}]
+        command = ['bulk_add_edge_by_ids', args, {'silent': silent}]
         return self._request([command])
 
-    def _add_entity_index(self, kind, *args):
+    def _add_entity_index(self, kind, *args, **filters):
         assert(kind in (KIND_EDGE, KIND_VERTEX))
-        command = ['add_node_index' if kind == KIND_VERTEX else 'add_edge_index', args, {}]
+        command = ['add_node_index' if kind == KIND_VERTEX else 'add_edge_index', args, filters]
         return self._request([command])
 
-    def add_node_index(self, *fields):
-        return self._add_entity_index(KIND_VERTEX, *fields)
+    def add_node_index(self, *fields, **filters):
+        return self._add_entity_index(KIND_VERTEX, *fields, **filters)
 
-    def add_edge_index(self, *fields):
-        return self._add_entity_index(KIND_EDGE, *fields)
+    def add_edge_index(self, *fields, **filters):
+        return self._add_entity_index(KIND_EDGE, *fields, **filters)
 
     def get_node_indexes(self):
         command = ['get_node_indexes', [], {}]
